#  Copyright 2021-present, the Recognai S.L. team.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from typing import TYPE_CHECKING

from datasets import Dataset, DatasetDict

from argilla.client.feedback.training.base import ArgillaTrainerSkeleton
from argilla.client.feedback.training.schemas.base import (
    TrainingTaskForQuestionAnswering,
    TrainingTaskForTextClassification,
)
from argilla.training.transformers import ArgillaTransformersTrainer as ArgillaTransformersTrainerV1
from argilla.utils.dependency import requires_dependencies

if TYPE_CHECKING:
    from argilla.client.feedback.integrations.huggingface.model_card import TransformersModelCardData


class ArgillaTransformersTrainer(ArgillaTransformersTrainerV1, ArgillaTrainerSkeleton):
    def __init__(self, *args, **kwargs):
        # init ArgillaTrainerSkeleton
        ArgillaTrainerSkeleton.__init__(self, *args, **kwargs)

        import torch
        from transformers import (
            AutoModelForQuestionAnswering,
            AutoModelForSequenceClassification,
            set_seed,
        )

        model = kwargs.get("model", None)
        self.trainer_model = model if model and not isinstance(model, str) else None
        self.trainer_tokenizer = kwargs.get("tokenizer", None)
        self.trainer_pipeline = None

        self.device = "cpu"
        if torch.backends.mps.is_available():
            self.device = "mps"
        elif torch.cuda.is_available():
            self.device = "cuda"

        if self._seed is None:
            self._seed = 42
        set_seed(self._seed)

        if self._model is None:
            self._model = "bert-base-cased"

        if isinstance(self._dataset, DatasetDict):
            self._train_dataset = self._dataset["train"]
            self._eval_dataset = self._dataset["test"]
        elif isinstance(self._dataset, Dataset):
            self._train_dataset = self._dataset
            self._eval_dataset = None
        else:
            raise NotImplementedError(f"We do not support {type(self._dataset)} yet.")

        if isinstance(self._task, TrainingTaskForTextClassification):
            self._model_class = AutoModelForSequenceClassification
        elif isinstance(self._task, TrainingTaskForQuestionAnswering):
            self._model_class = AutoModelForQuestionAnswering
        else:
            raise NotImplementedError(
                f"ArgillaTransformersTrainer does not support {self._task.__class__.__name__} yet."
            )

        self.init_training_args()

    def get_model_card_data(self, **card_data_kwargs) -> "TransformersModelCardData":
        """
        Generate the card data to be used for the `ArgillaModelCard`.

        Args:
            card_data_kwargs: Extra arguments provided by the user when creating the `ArgillaTrainer`.

        Returns:
            TransformersModelCardData: Container for the data to be written on the `ArgillaModelCard`.
        """
        from argilla.client.feedback.integrations.huggingface.model_card import TransformersModelCardData

        if not card_data_kwargs.get("tags"):
            if isinstance(self._task, TrainingTaskForTextClassification):
                tags = ["text-classification"]
            else:
                tags = ["question-answering"]

            card_data_kwargs.update({"tags": tags + ["transformers", "argilla"]})

        return TransformersModelCardData(
            model_id=self._model,
            task=self._task,
            update_config_kwargs=self.trainer_kwargs,
            **card_data_kwargs,
        )

    @requires_dependencies("huggingface_hub")
    def push_to_huggingface(self, repo_id: str, **kwargs) -> None:
        """Uploads the transformer model and tokenizer to [huggingface's model hub](https://huggingface.co/models).

        The full list of parameters can be seen at:
        [huggingface_hub](https://huggingface.co/docs/huggingface_hub/package_reference/mixins#huggingface_hub.ModelHubMixin.push_to_hub).

        Args:
            repo_id:
                The name of the repository you want to push your model and tokenizer to.
                It should contain your organization name when pushing to a given organization.

        Raises:
            NotImplementedError: If the model doesn't exist, meaning it hasn't been instantiated yet.
        """
        if not self.trainer_model:
            raise ValueError(
                "The model must be initialized prior to this point. You can either call `train` or `init_model`."
            )
        model_url = self.trainer_model.push_to_hub(repo_id, **kwargs)
        self._logger.info(f"Model pushed to: {model_url}")
        tokenizer_url = self.trainer_tokenizer.push_to_hub(repo_id, **kwargs)
        self._logger.info(f"Tokenizer pushed to: {tokenizer_url}")



#  Copyright 2021-present, the Recognai S.L. team.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import typer

from argilla.cli.callback import init_callback
from argilla.client.utils import server_info

app = typer.Typer(invoke_without_command=True)


@app.callback(help="Displays information about the Argilla client and server")
def info() -> None:
    from rich.console import Console
    from rich.markdown import Markdown

    from argilla._version import version
    from argilla.cli.rich import get_argilla_themed_panel

    init_callback()

    info = server_info()
    panel = get_argilla_themed_panel(
        Markdown(
            f"Connected to {info.url}\n"
            f"- **Client version:** {version}\n"
            f"- **Server version:** {info.version}\n"
            f"- **ElasticSearch version:** {info.elasticsearch_version}\n"
        ),
        title="Argilla Info",
        title_align="left",
    )

    Console().print(panel)


if __name__ == "__main__":
    app()



#  Copyright 2021-present, the Recognai S.L. team.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
import warnings
from abc import ABC, ABCMeta, abstractmethod
from typing import TYPE_CHECKING, Any, Generic, Iterable, List, Literal, Optional, Tuple, TypeVar, Union

from argilla.client.feedback.integrations.huggingface import HuggingFaceDatasetMixin
from argilla.client.feedback.schemas.records import FeedbackRecord, SortBy
from argilla.utils.dependency import requires_dependencies

if TYPE_CHECKING:
    from datasets import Dataset

    from argilla.client.feedback.schemas.types import (
        AllowedFieldTypes,
        AllowedMetadataPropertyTypes,
        AllowedQuestionTypes,
        AllowedVectorSettingsTypes,
    )


R = TypeVar("R", bound=FeedbackRecord)


class FeedbackDatasetBase(ABC, Generic[R], metaclass=ABCMeta):
    """Base class with shared functionality for `FeedbackDataset` and `RemoteFeedbackDataset`."""

    @property
    @abstractmethod
    def records(self) -> Iterable[R]:
        """Returns the records of the dataset."""
        pass

    @abstractmethod
    def update_records(self, records: Union[R, List[R]]) -> None:
        """Updates the records of the dataset.

        Args:
            records: the records to update the dataset with.

        Raises:
            ValueError: if the provided `records` are invalid.
        """
        pass

    @property
    @abstractmethod
    def guidelines(self) -> Optional[str]:
        """Returns the guidelines for annotating the dataset."""
        pass

    @property
    @abstractmethod
    def allow_extra_metadata(self) -> bool:
        """Returns whether if adding extra metadata to the records of the dataset is allowed"""
        pass

    def __get_property_by_name(self, item_name: str, iterable_items, item_type: str):
        for item in iterable_items:
            if item.name == item_name:
                return item
        warnings.warn(
            f"{item_type} with name='{item_name}' not found, available {item_type} names are:"
            f" {', '.join(item.name for item in iterable_items)}"
        )

    @property
    @abstractmethod
    def fields(self) -> List["AllowedFieldTypes"]:
        """Returns the fields that define the schema of the records in the dataset."""
        pass

    def field_by_name(self, name: str) -> Optional["AllowedFieldTypes"]:
        """Returns the field by name if it exists. Otherwise a `ValueError` is raised.

        Args:
            name: the name of the field to return.
        """
        return self.__get_property_by_name(name, self.fields, "field")

    @property
    @abstractmethod
    def questions(self) -> List["AllowedQuestionTypes"]:
        """Returns the questions that will be used to annotate the dataset."""
        pass

    def question_by_name(self, name: str) -> Optional["AllowedQuestionTypes"]:
        """Returns the question by name if it exists.

        Args:
            name: the name of the question to return.
        """
        return self.__get_property_by_name(name, self.questions, "question")

    @property
    @abstractmethod
    def metadata_properties(
        self,
    ) -> List["AllowedMetadataPropertyTypes"]:
        """Returns the metadata properties that will be indexed and could be used to filter the dataset."""
        pass

    def metadata_property_by_name(self, name: str) -> Optional["AllowedMetadataPropertyTypes"]:
        """Returns the metadata property by name if it exists.

        Args:
            name: the name of the metadata property to return.
        """
        return self.__get_property_by_name(name, self.metadata_properties, "metadata property")

    @property
    @abstractmethod
    def vectors_settings(self) -> List["AllowedVectorSettingsTypes"]:
        """Returns the vector settings that will be used to index the dataset."""
        pass

    def vector_settings_by_name(self, name: str) -> Optional["AllowedVectorSettingsTypes"]:
        """Returns the vector settings by name if it exists.

        Args:
            name: the name of the vector settings to return.

        Raises:
            KeyError: if the vector settings with the given name does not exist.
        """
        return self.__get_property_by_name(name, self.vectors_settings, "vector settings")

    @abstractmethod
    def sort_by(self, sort: List[SortBy]) -> "FeedbackDatasetBase":
        """Sorts the records in the dataset by the given field."""
        pass

    def _unique_metadata_property(self, metadata_property: "AllowedMetadataPropertyTypes") -> None:
        """Checks whether the provided `metadata_property` already exists in the dataset.

        Args:
            metadata_property: the metadata property to validate.

        Raises:
            ValueError: if the `metadata_property` already exists in the dataset.
        """
        existing_metadata_properties = self.metadata_properties
        if existing_metadata_properties:
            existing_metadata_property_names = [
                metadata_property.name for metadata_property in existing_metadata_properties
            ]
            if metadata_property.name in existing_metadata_property_names:
                raise ValueError(
                    f"Invalid `metadata_property={metadata_property.name}` provided as it already exists. Current"
                    f" `metadata_properties` are: {', '.join(existing_metadata_property_names)}"
                )

    @requires_dependencies("datasets")
    def format_as(self, format: Literal["datasets"]) -> "Dataset":
        """Formats the `FeedbackDataset` as a `datasets.Dataset` object.

        Args:
            format: the format to use to format the `FeedbackDataset`. Currently supported formats are:
                `datasets`.

        Returns:
            The `FeedbackDataset.records` formatted as a `datasets.Dataset` object.

        Raises:
            ValueError: if the provided format is not supported.

        Examples:
            >>> import argilla as rg
            >>> rg.init(...)
            >>> dataset = rg.FeedbackDataset.from_argilla(name="my-dataset")
            >>> huggingface_dataset = dataset.format_as("datasets")
        """
        if format == "datasets":
            return HuggingFaceDatasetMixin._huggingface_format(self)
        raise ValueError(f"Unsupported format '{format}'.")

    @abstractmethod
    def add_records(self, *args, **kwargs) -> None:
        """Adds the given records to the `FeedbackDataset`."""
        pass

    @abstractmethod
    def pull(self):
        """Pulls the dataset from Argilla and returns a local instance of it."""
        pass

    @abstractmethod
    def filter_by(self, *args, **kwargs):
        """Filters the current `FeedbackDataset`."""
        pass

    @abstractmethod
    def delete(self):
        """Deletes the `FeedbackDataset` from Argilla."""
        pass

    @abstractmethod
    def prepare_for_training(self, *args, **kwargs) -> Any:
        """Prepares the `FeedbackDataset` for training by creating the training."""
        pass

    @abstractmethod
    def push_to_argilla(self, *args, **kwargs) -> "FeedbackDatasetBase":
        """Pushes the `FeedbackDataset` to Argilla."""
        pass

    @abstractmethod
    def add_metadata_property(self, *args, **kwargs):
        """Adds a new `metadata_property` to the current `FeedbackDataset`."""
        pass

    @abstractmethod
    def update_metadata_properties(self, *args, **kwargs):
        """Updates the `metadata_properties` of the current `FeedbackDataset`."""
        pass

    @abstractmethod
    def delete_metadata_properties(self, *args, **kwargs):
        """Deletes a list of `metadata_properties` from the current `FeedbackDataset`."""
        pass

    @abstractmethod
    def add_vector_settings(self, *args, **kwargs):
        """Adds a new `vector_settings` to the current `FeedbackDataset`."""
        pass

    @abstractmethod
    def update_vectors_settings(self, *args, **kwargs):
        """Updates a list of `vector_settings` from the current `FeedbackDataset`."""
        pass

    @abstractmethod
    def delete_vectors_settings(self, *args, **kwargs):
        """Deletes a list of `vector_settings` from the current `FeedbackDataset`."""
        pass

    @abstractmethod
    def push_to_huggingface(self, repo_id, generate_card, *args, **kwargs):
        """Pushes the current `FeedbackDataset` to HuggingFace Hub.

        Note:
            The records from the `RemoteFeedbackDataset` are being pulled before pushing,
            to ensure that there's no missmatch while uploading those as those are lazily fetched.

        Args:
            repo_id: the ID of the HuggingFace repo to push the dataset to.
            generate_card: whether to generate a dataset card or not. Defaults to `True`.
        """
        pass

    @abstractmethod
    def find_similar_records(
        self,
        vector_name: str,
        value: Optional[List[float]] = None,
        record: Optional[R] = None,
        max_results: int = 50,
    ) -> List[Tuple[R, float]]:
        """Finds similar records to the given `record` or `value` for the given `vector_name`.

        Args:
            vector_name: a vector name to use for searching by similarity.
            value: an optional vector value to be used for searching by similarity.
            record: an optional record to be used for searching by similarity.
            max_results: the maximum number of results for the search.

        Returns:
            A list of tuples with each tuple including a record and a similarity score.
        """
        pass



#  Copyright 2021-present, the Recognai S.L. team.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from typing import TYPE_CHECKING

from argilla.client.feedback.schemas.fields import TextField
from argilla.client.feedback.schemas.remote.shared import RemoteSchema

if TYPE_CHECKING:
    from argilla.client.sdk.v1.datasets.models import FeedbackFieldModel


class RemoteTextField(TextField, RemoteSchema):
    def to_local(self) -> TextField:
        return TextField(
            name=self.name,
            title=self.title,
            required=self.required,
            use_markdown=self.use_markdown,
        )

    @classmethod
    def from_api(cls, payload: "FeedbackFieldModel") -> "RemoteTextField":
        return RemoteTextField(
            id=payload.id,
            name=payload.name,
            title=payload.title,
            required=payload.required,
            use_markdown=payload.settings["use_markdown"],
        )



#  Copyright 2021-present, the Recognai S.L. team.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from typing import TYPE_CHECKING, Dict, List, Union

from argilla.client.feedback.schemas.questions import (
    LabelQuestion,
    MultiLabelQuestion,
    RankingQuestion,
    RatingQuestion,
    TextQuestion,
)
from argilla.client.feedback.schemas.remote.shared import RemoteSchema

if TYPE_CHECKING:
    from argilla.client.sdk.v1.datasets.models import FeedbackQuestionModel


class RemoteTextQuestion(TextQuestion, RemoteSchema):
    def to_local(self) -> TextQuestion:
        return TextQuestion(
            name=self.name,
            title=self.title,
            description=self.description,
            required=self.required,
            use_markdown=self.use_markdown,
        )

    @classmethod
    def from_api(cls, payload: "FeedbackQuestionModel") -> "RemoteTextQuestion":
        return RemoteTextQuestion(
            id=payload.id,
            name=payload.name,
            title=payload.title,
            required=payload.required,
            use_markdown=payload.settings["use_markdown"],
        )


class RemoteRatingQuestion(RatingQuestion, RemoteSchema):
    def to_local(self) -> RatingQuestion:
        return RatingQuestion(
            name=self.name,
            title=self.title,
            description=self.description,
            required=self.required,
            values=self.values,
        )

    @classmethod
    def from_api(cls, payload: "FeedbackQuestionModel") -> "RemoteRatingQuestion":
        return RemoteRatingQuestion(
            id=payload.id,
            name=payload.name,
            title=payload.title,
            required=payload.required,
            values=[option["value"] for option in payload.settings["options"]],
        )


def _parse_options_from_api(payload: "FeedbackQuestionModel") -> Union[List[str], Dict[str, str]]:
    if all([label["value"] == label["text"] for label in payload.settings["options"]]):
        return [label["value"] for label in payload.settings["options"]]
    else:
        return {label["value"]: label["text"] for label in payload.settings["options"]}


class RemoteLabelQuestion(LabelQuestion, RemoteSchema):
    def to_local(self) -> LabelQuestion:
        return LabelQuestion(
            name=self.name,
            title=self.title,
            description=self.description,
            required=self.required,
            labels=self.labels,
            visible_labels=self.visible_labels,
        )

    @classmethod
    def from_api(cls, payload: "FeedbackQuestionModel") -> "RemoteLabelQuestion":
        return RemoteLabelQuestion(
            id=payload.id,
            name=payload.name,
            title=payload.title,
            required=payload.required,
            labels=_parse_options_from_api(payload),
            visible_labels=payload.settings["visible_options"],
        )


class RemoteMultiLabelQuestion(MultiLabelQuestion, RemoteSchema):
    def to_local(self) -> MultiLabelQuestion:
        return MultiLabelQuestion(
            name=self.name,
            title=self.title,
            description=self.description,
            required=self.required,
            labels=self.labels,
            visible_labels=self.visible_labels,
        )

    @classmethod
    def from_api(cls, payload: "FeedbackQuestionModel") -> "RemoteMultiLabelQuestion":
        return RemoteMultiLabelQuestion(
            id=payload.id,
            name=payload.name,
            title=payload.title,
            required=payload.required,
            labels=_parse_options_from_api(payload),
            visible_labels=payload.settings["visible_options"],
        )


class RemoteRankingQuestion(RankingQuestion, RemoteSchema):
    def to_local(self) -> RankingQuestion:
        return RankingQuestion(
            name=self.name,
            title=self.title,
            description=self.description,
            required=self.required,
            values=self.values,
        )

    @classmethod
    def from_api(cls, payload: "FeedbackQuestionModel") -> "RemoteRankingQuestion":
        return RemoteRankingQuestion(
            id=payload.id,
            name=payload.name,
            title=payload.title,
            required=payload.required,
            values=_parse_options_from_api(payload),
        )



#  Copyright 2021-present, the Recognai S.L. team.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from typing import List, Optional, Union

from argilla.client.feedback.schemas import (
    LabelQuestion,
    MultiLabelQuestion,
    RankingQuestion,
    RatingQuestion,
    TextField,
    TextQuestion,
)
from argilla.client.feedback.unification import (
    LabelQuestionUnification,
    MultiLabelQuestionUnification,
    RankingQuestionUnification,
    RatingQuestionUnification,
)
from argilla.pydantic_v1 import BaseModel


class TextClassificationDefaults(BaseModel):
    text: Optional[TextField] = None
    label: Optional[
        Union[
            RatingQuestion,
            LabelQuestion,
            MultiLabelQuestion,
            RankingQuestion,
            LabelQuestionUnification,
            MultiLabelQuestionUnification,
            RankingQuestionUnification,
            RatingQuestionUnification,
        ]
    ] = None


class QuestionAnsweringDefaults(BaseModel):
    question: TextField
    context: TextField
    answer: TextQuestion


class SentenceSimilarityDefaults(BaseModel):
    texts: Optional[List[TextField]] = None
    label: Optional[Union[LabelQuestion, RatingQuestion, LabelQuestionUnification, RankingQuestionUnification]] = None



#  Copyright 2021-present, the Recognai S.L. team.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import json
import logging
import tempfile
import warnings
from typing import TYPE_CHECKING, Any, Optional, Type, Union

from packaging.version import parse as parse_version
from tqdm import trange

from argilla.client.feedback.constants import FIELD_TYPE_TO_PYTHON_TYPE
from argilla.client.feedback.schemas.enums import QuestionTypes
from argilla.client.feedback.schemas.records import FeedbackRecord
from argilla.client.feedback.schemas.remote.records import RemoteFeedbackRecord
from argilla.utils.dependency import requires_dependencies

if TYPE_CHECKING:
    from datasets import Dataset

    from argilla.client.feedback.dataset.local.dataset import FeedbackDataset
    from argilla.client.feedback.dataset.remote.dataset import RemoteFeedbackDataset

_LOGGER = logging.getLogger(__name__)


class HuggingFaceDatasetMixin:
    @staticmethod
    @requires_dependencies("datasets")
    def _huggingface_format(dataset: Union["FeedbackDataset", "RemoteFeedbackDataset"]) -> "Dataset":
        """Formats either a `FeedbackDataset` or a `RemoteFeedbackDataset` as a `datasets.Dataset` object.

        Args:
            dataset: The `FeedbackDataset` or `RemoteFeedbackDataset` to format as `datasets.Dataset`.

        Returns:
            The records from the dataset formatted as a `datasets.Dataset` object, including the fields,
            questions, and metadata_properties formatted as `datasets.Features`.

        Examples:
            >>> from argilla.client.feedback.integrations.dataset import HuggingFaceDatasetMixin
            >>> dataset = FeedbackDataset(...) or RemoteFeedbackDataset(...)
            >>> huggingface_dataset = HuggingFaceDatasetMixin._huggingface_format(dataset)
        """
        from datasets import Dataset, Features, Sequence, Value

        hf_dataset, hf_features = {}, {}

        for field in dataset.fields:
            if field.type not in FIELD_TYPE_TO_PYTHON_TYPE.keys():
                raise ValueError(
                    f"Field {field.name} has an unsupported type: {field.type}, for the moment"
                    f" only the following types are supported: {list(FIELD_TYPE_TO_PYTHON_TYPE.keys())}"
                )
            hf_features[field.name] = Value(dtype="string", id="field")
            if field.name not in hf_dataset:
                hf_dataset[field.name] = []

        for question in dataset.questions:
            if question.type in [QuestionTypes.text, QuestionTypes.label_selection]:
                value = Value(dtype="string", id="question")
            elif question.type == QuestionTypes.rating:
                value = Value(dtype="int32", id="question")
            elif question.type == QuestionTypes.ranking:
                value = Sequence({"rank": Value(dtype="uint8"), "value": Value(dtype="string")}, id="question")
            elif question.type in QuestionTypes.multi_label_selection:
                value = Sequence(Value(dtype="string"), id="question")
            else:
                raise ValueError(
                    f"Question {question.name} is of type `{question.type}`,"
                    " for the moment only the following question types are supported:"
                    f" `{'`, `'.join([arg.value for arg in QuestionTypes])}`."
                )

            hf_features[question.name] = [
                {
                    "user_id": Value(dtype="string", id="question"),
                    "value": value,
                    "status": Value(dtype="string", id="question"),
                }
            ]
            if question.name not in hf_dataset:
                hf_dataset[question.name] = []

            value.id = "suggestion"
            hf_features[f"{question.name}-suggestion"] = value
            if f"{question.name}-suggestion" not in hf_dataset:
                hf_dataset[f"{question.name}-suggestion"] = []

            hf_features[f"{question.name}-suggestion-metadata"] = {
                "type": Value(dtype="string", id="suggestion-metadata"),
                "score": Value(dtype="float32", id="suggestion-metadata"),
                "agent": Value(dtype="string", id="suggestion-metadata"),
            }
            if f"{question.name}-suggestion-metadata" not in hf_dataset:
                hf_dataset[f"{question.name}-suggestion-metadata"] = []

        hf_features["external_id"] = Value(dtype="string", id="external_id")
        hf_dataset["external_id"] = []

        hf_features["metadata"] = Value(dtype="string", id="metadata")
        hf_dataset["metadata"] = []

        vectors_settings = dataset.vectors_settings
        if vectors_settings:
            hf_features["vectors"] = {}
            for vector_settings in vectors_settings:
                hf_features["vectors"].update({vector_settings.name: Sequence(Value(dtype="float32"), id="vectors")})
            hf_dataset["vectors"] = []

        for record in dataset.records:
            for field in dataset.fields:
                hf_dataset[field.name].append(record.fields.get(field.name, None))
            for question in dataset.questions:
                if not record.responses:
                    hf_dataset[question.name].append([])
                else:
                    responses = []
                    for response in record.responses:
                        if question.name not in response.values:
                            continue
                        formatted_response = {
                            "user_id": response.user_id,
                            "value": None,
                            "status": response.status.value if hasattr(response.status, "value") else response.status,
                        }
                        if question.type == QuestionTypes.ranking:
                            value = [r.dict() for r in response.values[question.name].value]
                        else:
                            value = response.values[question.name].value
                        formatted_response["value"] = value
                        responses.append(formatted_response)
                    hf_dataset[question.name].append(responses)

                suggestion_value, suggestion_metadata = None, {"type": None, "score": None, "agent": None}
                if record.suggestions:
                    for suggestion in record.suggestions:
                        if question.name == suggestion.question_name:
                            suggestion_value = suggestion.value
                            suggestion_metadata = {
                                "type": suggestion.type,
                                "score": suggestion.score,
                                "agent": suggestion.agent,
                            }
                            break
                hf_dataset[f"{question.name}-suggestion"].append(suggestion_value)
                hf_dataset[f"{question.name}-suggestion-metadata"].append(suggestion_metadata)

            hf_dataset["metadata"].append(json.dumps(record.metadata) if record.metadata else {})
            hf_dataset["external_id"].append(record.external_id or None)

            if vectors_settings:
                vectors = {}
                for vector_settings in vectors_settings:
                    vectors.update({vector_settings.name: record.vectors.get(vector_settings.name, None)})
                hf_dataset["vectors"].append(vectors)

        return Dataset.from_dict(hf_dataset, features=Features(hf_features))

    @requires_dependencies(["huggingface_hub", "datasets"])
    def push_to_huggingface(
        self: "FeedbackDataset",
        repo_id: str,
        generate_card: Optional[bool] = True,
        *args,
        **kwargs,
    ) -> None:
        """Pushes the `FeedbackDataset` to the Hugging Face Hub. If the dataset has been previously pushed to the
        Hugging Face Hub, it will be updated instead. Note that some params as `private` have no effect at all
        when a dataset is previously uploaded to the Hugging Face Hub.

        Args:
            dataset: the `FeedbackDataset` to push to the Hugging Face Hub.
            repo_id: the ID of the Hugging Face Hub repo to push the `FeedbackDataset` to.
            generate_card: whether to generate a dataset card for the `FeedbackDataset` in the Hugging Face Hub. Defaults
                to `True`.
            *args: the args to pass to `datasets.Dataset.push_to_hub`.
            **kwargs: the kwargs to pass to `datasets.Dataset.push_to_hub`.
        """
        import huggingface_hub
        from huggingface_hub import HfApi

        # https://github.com/argilla-io/argilla/issues/3468
        from argilla.client.feedback.config import DatasetConfig

        if parse_version(huggingface_hub.__version__) < parse_version("0.14.0"):
            _LOGGER.warning(
                "Recommended `huggingface_hub` version is 0.14.0 or higher, and you have"
                f" {huggingface_hub.__version__}, so in case you have any issue when pushing the dataset to the"
                " Hugging Face Hub upgrade it as `pip install huggingface_hub --upgrade`."
            )

        if len(self) < 1:
            raise ValueError(
                "Cannot push an empty `rg.FeedbackDataset` to the Hugging Face Hub, please make sure to add at"
                " least one record, via the method `add_records`."
            )

        hfds = self.format_as("datasets")
        hfds.push_to_hub(repo_id, *args, **kwargs)

        with tempfile.NamedTemporaryFile(mode="w", delete=False) as f:
            f.write(
                DatasetConfig(
                    fields=self.fields,
                    questions=self.questions,
                    guidelines=self.guidelines,
                    metadata_properties=self.metadata_properties or None,
                    allow_extra_metadata=self.allow_extra_metadata,
                    vectors_settings=self.vectors_settings or None,
                ).to_yaml()
            )
            f.flush()

            HfApi().upload_file(
                path_or_fileobj=f.name,
                path_in_repo="argilla.yaml",
                repo_id=repo_id,
                repo_type="dataset",
                token=kwargs.get("token"),
            )

        if generate_card:
            from huggingface_hub import DatasetCardData

            from argilla.client.feedback.integrations.huggingface.card import (
                ArgillaDatasetCard,
                size_categories_parser,
            )

            sample_argilla_record = self.records[0]
            sample_argilla_record = (
                sample_argilla_record.to_local()
                if isinstance(sample_argilla_record, RemoteFeedbackRecord)
                else sample_argilla_record
            )
            sample_huggingface_record = hfds[0]

            card = ArgillaDatasetCard.from_template(
                card_data=DatasetCardData(
                    size_categories=size_categories_parser(len(self.records)),
                    tags=["rlfh", "argilla", "human-feedback"],
                ),
                repo_id=repo_id,
                argilla_fields=self.fields,
                argilla_questions=self.questions,
                argilla_guidelines=self.guidelines or None,
                argilla_vectors_settings=self.vectors_settings or None,
                argilla_metadata_properties=self.metadata_properties,
                argilla_record=json.loads(sample_argilla_record.json()),
                huggingface_record=sample_huggingface_record,
            )
            card.push_to_hub(repo_id, repo_type="dataset", token=kwargs.get("token"))

    @classmethod
    @requires_dependencies(["huggingface_hub", "datasets"])
    def from_huggingface(
        cls: Type["FeedbackDataset"], repo_id: str, show_progress: bool = True, *args: Any, **kwargs: Any
    ) -> "FeedbackDataset":
        """Loads a `FeedbackDataset` from the Hugging Face Hub.

        Args:
            repo_id: the ID of the Hugging Face Hub repo to load the `FeedbackDataset` from.
            *args: the args to pass to `datasets.Dataset.load_from_hub`.
            **kwargs: the kwargs to pass to `datasets.Dataset.load_from_hub`.

        Returns:
            A `FeedbackDataset` loaded from the Hugging Face Hub.
        """
        import huggingface_hub
        from datasets import DatasetDict, load_dataset
        from huggingface_hub import hf_hub_download
        from huggingface_hub.utils import EntryNotFoundError

        # https://github.com/argilla-io/argilla/issues/3468
        from argilla.client.feedback.config import (
            DatasetConfig,
            DeprecatedDatasetConfig,
        )

        if parse_version(huggingface_hub.__version__) < parse_version("0.14.0"):
            _LOGGER.warning(
                "Recommended `huggingface_hub` version is 0.14.0 or higher, and you have"
                f" {huggingface_hub.__version__}, so in case you have any issue when pushing the dataset to the"
                " Hugging Face Hub upgrade it as `pip install huggingface_hub --upgrade`."
            )

        if "token" in kwargs:
            auth = kwargs.pop("token")
        elif "use_auth_token" in kwargs:
            auth = kwargs.pop("use_auth_token")
        else:
            auth = None

        hub_auth = (
            {"use_auth_token": auth}
            if parse_version(huggingface_hub.__version__) < parse_version("0.11.0")
            else {"token": auth}
        )

        try:
            config_path = hf_hub_download(
                repo_id=repo_id,
                filename="argilla.yaml",
                repo_type="dataset",
                **hub_auth,
            )
            with open(config_path, "r") as f:
                config = DatasetConfig.from_yaml(f.read())
                dataset = cls(
                    fields=config.fields,
                    questions=config.questions,
                    guidelines=config.guidelines,
                    metadata_properties=config.metadata_properties,
                    allow_extra_metadata=config.allow_extra_metadata,
                    vectors_settings=config.vectors_settings,
                )
        except EntryNotFoundError:
            # TODO(alvarobartt): here for backwards compatibility, last used in 1.12.0
            warnings.warn(
                "No `argilla.yaml` file found in the Hugging Face Hub repository, which"
                " means that the `DatasetConfig` was dumped using Argilla 1.12.0 or"
                " lower, and the `argilla.yaml` file was not generated. Please consider"
                " re-dumping the `DatasetConfig` using Argilla 1.13.0 or higher, or"
                " manually create the `argilla.yaml` file in the Hugging Face Hub.",
                UserWarning,
            )
            config_path = hf_hub_download(
                repo_id=repo_id,
                filename="argilla.cfg",
                repo_type="dataset",
                **hub_auth,
            )
            with open(config_path, "r") as f:
                config = DeprecatedDatasetConfig.from_json(f.read())
                dataset = cls(fields=config.fields, questions=config.questions, guidelines=config.guidelines)
        except Exception as e:
            raise FileNotFoundError(
                "Neither `argilla.yaml` nor `argilla.cfg` files were found in the"
                " Hugging Face Hub repository. Please make sure to dump the `DatasetConfig`"
                " using `FeedbackDataset.push_to_huggingface` to automatically upload"
                " the `DatasetConfig` as `argilla.yaml` to the Hugging Face Hub."
            ) from e

        hfds = load_dataset(repo_id, token=auth, *args, **kwargs)  # use_auth_token is deprecated
        if isinstance(hfds, DatasetDict) and "split" not in kwargs:
            if len(hfds.keys()) > 1:
                raise ValueError(
                    "Only one dataset can be loaded at a time, use `split` to select a split, available splits"
                    f" are: {', '.join(hfds.keys())}."
                )
            hfds = hfds[list(hfds.keys())[0]]

        records = []
        for index in trange(0, len(hfds), desc="Parsing records", disable=not show_progress):
            responses = {}
            suggestions = []
            user_without_id = False
            for question in dataset.questions:
                if hfds[index][question.name] is not None and len(hfds[index][question.name]) > 0:
                    if (
                        len(
                            [None for response in hfds[index][question.name] if response["user_id"] is None]
                            if isinstance(hfds[index][question.name], list)
                            else [None for user_id in hfds[index][question.name]["user_id"] if user_id is None]
                        )
                        > 1
                    ):
                        warnings.warn(
                            "Found more than one user without ID in the dataset, so just the"
                            " responses for the first user without ID will be used, the rest"
                            " will be discarded."
                        )

                    # Here for backwards compatibility
                    original_responses = []
                    if isinstance(hfds[index][question.name], list):
                        original_responses = hfds[index][question.name]
                    else:
                        for user_id, value, status in zip(
                            hfds[index][question.name]["user_id"],
                            hfds[index][question.name]["value"],
                            hfds[index][question.name]["status"],
                        ):
                            original_responses.append({"user_id": user_id, "value": value, "status": status})

                    user_without_id_response = False
                    for response in original_responses:
                        if user_without_id_response:
                            continue
                        user_id = response["user_id"]
                        status = response["status"]
                        if user_id is None:
                            if not user_without_id:
                                user_without_id = True
                                responses["user_without_id"] = {
                                    "user_id": user_id,
                                    "status": status,
                                    "values": {},
                                }
                                user_without_id_response = True
                        if user_id is not None and user_id not in responses:
                            responses[user_id] = {
                                "user_id": user_id,
                                "status": status,
                                "values": {},
                            }
                        value = response["value"]
                        if value is not None:
                            if question.type == QuestionTypes.ranking:
                                value = [{"rank": r, "value": v} for r, v in zip(value["rank"], value["value"])]
                            responses[user_id or "user_without_id"]["values"].update({question.name: {"value": value}})

                # First if-condition is here for backwards compatibility
                if (
                    f"{question.name}-suggestion" in hfds[index]
                    and hfds[index][f"{question.name}-suggestion"] is not None
                ):
                    suggestion = {
                        "question_name": question.name,
                        "value": hfds[index][f"{question.name}-suggestion"],
                    }
                    if hfds[index][f"{question.name}-suggestion-metadata"] is not None:
                        suggestion.update(hfds[index][f"{question.name}-suggestion-metadata"])
                    suggestions.append(suggestion)

            metadata = None
            if "metadata" in hfds[index] and hfds[index]["metadata"] is not None:
                metadata = json.loads(hfds[index]["metadata"])

            vectors = {}
            if "vectors" in hfds[index] and hfds[index]["vectors"]:
                for vector_settings in dataset.vectors_settings:
                    if hfds[index]["vectors"].get(vector_settings.name, None) is not None:
                        vectors.update({vector_settings.name: hfds[index]["vectors"][vector_settings.name]})

            records.append(
                FeedbackRecord(
                    fields={field.name: hfds[index][field.name] for field in dataset.fields},
                    metadata=metadata or {},
                    responses=list(responses.values()) or [],
                    suggestions=[suggestion for suggestion in suggestions if suggestion["value"] is not None] or [],
                    vectors=vectors or {},
                    external_id=hfds[index]["external_id"],
                )
            )
        del hfds

        dataset.add_records(records)

        return dataset



#  coding=utf-8
#  Copyright 2021-present, the Recognai S.L. team.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
import logging
import re
import warnings
from typing import Any, List, Optional, Union

import numpy as np
import pandas as pd
from rich.progress import Progress

from argilla.client.feedback.dataset.local.dataset import FeedbackDataset
from argilla.client.feedback.dataset.remote.dataset import RemoteFeedbackDataset
from argilla.client.feedback.schemas.metadata import (
    FloatMetadataProperty,
    IntegerMetadataProperty,
    TermsMetadataProperty,
)
from argilla.client.feedback.schemas.records import FeedbackRecord
from argilla.client.feedback.schemas.remote.records import RemoteFeedbackRecord
from argilla.utils.dependency import require_dependencies

_LOGGER = logging.getLogger(__name__)
_LOGGER.setLevel(logging.INFO)


class TextDescriptivesExtractor:
    """This class extracts a number of basic text descriptives from FeedbackDataset
    records using the TextDescriptives library and adds them as record metadata."""

    def __init__(
        self,
        model: str = "en",
        metrics: Optional[List[str]] = None,
        visible_for_annotators: Optional[bool] = True,
        show_progress: Optional[bool] = True,
    ):
        """
        Initialize a new TextDescriptivesExtractor object.

        Args:
            model (str): The language of the model to use for text descriptives.
            metrics (Optional[List[str]]): A list of metrics to extract
                [“descriptive_stats”, “readability”, “dependency_distance”, “pos_proportions”, “coherence”, “quality”, “information_theory”].
                If None, all metrics will be extracted.
            visible_for_annotators (bool): Whether the extracted metrics should be visible to annotators.
            show_progress (bool): Whether to show a progress bar when extracting metrics.

        Examples:
        >>> import argilla as rg
        >>> from argilla.client.feedback.integrations.textdescriptives import TextDescriptivesExtractor
        >>> ds = rg.FeedbackDataset(...)
        >>> tde = TextDescriptivesExtractor()
        >>> updated_ds = tde.update_dataset(ds)
        >>> updated_records = tde.update_records(ds.records)
        """
        require_dependencies("textdescriptives")
        if model == "en" or model is None:
            model = "en_core_web_md"
            warnings.warn("Using 'en_core_web_md' as default spaCy model for English language.")
        self.model = model
        self.metrics = metrics
        self.__basic_metrics = [
            "n_tokens",
            "n_unique_tokens",
            "n_sentences",
            "perplexity",
            "entropy",
            "flesch_reading_ease",
        ]
        self.visible_for_annotators = visible_for_annotators
        self.show_progress = show_progress

    def _extract_metrics_for_single_field(
        self,
        records: List[Union[FeedbackRecord, RemoteFeedbackRecord]],
        field: str,
        overwrite: bool = False,
        basic_metrics: Optional[List[str]] = None,
    ) -> Optional[pd.DataFrame]:
        """
        Extract text descriptives metrics for a single field from a list of feedback records
        using the TextDescriptives library.

        Args:
            records (List[Union[FeedbackRecord, RemoteFeedbackRecord]]): A list of FeedbackDataset or RemoteFeedbackDataset records.
            field (str): The name of the field to extract metrics for.
            basic_metrics (Optional[List[str]]): A list of basic metrics to extract. If None, all metrics will be extracted.

        Returns:
            Optional[pd.DataFrame]: A dataframe containing the text descriptives metrics for the field, or None if the field is empty.
        """
        import textdescriptives as td

        def _get_metrics_from_texts(texts: List[str], basic_metrics):
            # If language is english, the default spacy model is used (to avoid warning message)
            field_metrics = td.extract_metrics(text=texts, spacy_model=self.model, metrics=self.metrics)
            # Drop text column
            field_metrics = field_metrics.drop("text", axis=1)
            # If basic metrics is None, use all basic metrics
            if basic_metrics is None and self.metrics is None:
                basic_metrics = self.__basic_metrics
                field_metrics = field_metrics.loc[:, basic_metrics]
            # Convert any None values to NaNs
            field_metrics = field_metrics.fillna(value=np.nan)
            # Select all column names that contain ONLY NaNs
            nan_columns = field_metrics.columns[field_metrics.isnull().all()].tolist()
            if nan_columns:
                warnings.warn(f"The following columns for {field} contain only NaN values: {nan_columns}")
            # Concatenate field name with the metric name
            field_metrics.columns = [f"{field}_{metric}" for metric in field_metrics.columns]
            return field_metrics

        idx_new = []
        idx_present = []
        data_present: List[dict] = []
        field_text = []
        columns: list[str] = _get_metrics_from_texts(["test"], basic_metrics=basic_metrics).columns.tolist()
        for idx, record in enumerate(records):
            if all([field in record.metadata for field in columns]) and not overwrite:
                idx_present.append(idx)
                metadata_relevant: dict[str, Any] = {key: record.metadata[key] for key in columns}
                data_present.append(metadata_relevant)
            elif record.fields.get(field):
                idx_new.append(idx)
                field_text.append(record.fields[field])
            else:
                idx_present.append(idx)
                data_present.append({key: np.nan for key in columns})

        # If the field is empty, skip it
        if not field_text:
            return None

        # Extract metrics for the field
        field_metrics_present = pd.DataFrame(data_present, index=idx_present)
        field_metrics_new = _get_metrics_from_texts(field_text, basic_metrics)
        field_metrics_new.index = idx_new

        # Combine metrics for the field
        field_metrics_combined = pd.concat([field_metrics_present, field_metrics_new])
        field_metrics_combined = field_metrics_combined.sort_index()

        return field_metrics_combined

    def _extract_metrics_for_all_fields(
        self,
        records: List[Union[FeedbackRecord, RemoteFeedbackRecord]],
        fields: List[str],
        overwrite: bool,
    ) -> pd.DataFrame:
        """
        Extract text descriptives metrics for all named fields from a list of feedback records
        using the TextDescriptives library.
        Args:
            records (List[Union[FeedbackRecord, RemoteFeedbackRecord]]): A list of FeedbackDataset or RemoteFeedbackDataset records.
            fields (List[str]): A list of fields to extract metrics for. If None, extract metrics for all fields.
            overwrite (bool): Whether to overwrite existing metadata properties with the same name. Defaults to False.
        Returns:
            pd.DataFrame: A dataframe containing the text descriptives metrics for each record and field.
        """
        # Extract all metrics for each field
        field_metrics = {
            field: self._extract_metrics_for_single_field(records=records, field=field, overwrite=overwrite)
            for field in fields
        }
        field_metrics = {field: metrics for field, metrics in field_metrics.items() if metrics is not None}
        # If there is only one field, return the metrics for that field directly
        if len(field_metrics) == 1:
            return list(field_metrics.values())[0]
        else:
            # If there are multiple fields, combine metrics for each field into a single dataframe
            final_metrics = pd.concat(field_metrics, axis=1, keys=field_metrics.keys())
            final_metrics.columns = final_metrics.columns.droplevel(0)
        return final_metrics

    def _cast_to_python_types(self, df: pd.DataFrame) -> pd.DataFrame:
        """
        Convert integer, boolean and floats columns in a dataframe
        to Python native types.

        Args:
            df (pd.DataFrame): The text descriptives dataframe.

        Returns:
            pd.DataFrame: The text descriptives dataframe with integer and boolean columns cast to Python native types.
        """
        # Select columns by data type
        int_cols = df.select_dtypes(include=["int64"]).columns
        bool_cols = df.select_dtypes(include=["boolean"]).columns
        float_cols = df.select_dtypes(include=["float64"]).columns
        # Cast integer columns to Python's native int type
        df[int_cols] = df[int_cols].astype(int)
        # Cast boolean columns to Python's native str type
        df[bool_cols] = df[bool_cols].astype(str)
        # Cast float columns to Python's native float type and round to 2 decimal places
        df[float_cols] = df[float_cols].astype(float).round(2)
        return df

    def _clean_column_name(self, col_name: str) -> str:
        """
        Clean the column name of a dataframe to fit a specific regex pattern.
        Args:
            col_name (str): A column name.
        Returns:
            str: A column name that fits the regex pattern.
        """
        col_name = col_name.lower()  # Convert to lowercase
        col_name = re.sub(r"[^a-z0-9_]", "_", col_name)  # Replace non-alphanumeric characters with underscores
        return col_name

    def _create_metadata_properties(self, df: pd.DataFrame) -> List:
        """
        Generate metadata properties based on dataframe columns and data types.

        Args:
            df (pd.DataFrame): The text descriptives dataframe.

        Returns:
            List: A list of metadata properties.
        """
        properties = []
        for col, dtype in df.dtypes.items():
            name = col
            title = name.replace("_", " ").title()
            if dtype in ["object", "bool"]:
                prop = TermsMetadataProperty(
                    name=name,
                    title=title,
                    visible_for_annotators=self.visible_for_annotators,
                    values=df[col].unique().tolist(),
                )
            elif dtype in ["int64", "int32", "int16", "int8", "int128", "int256", "int"]:
                prop = IntegerMetadataProperty(
                    name=name, title=title, visible_for_annotators=self.visible_for_annotators
                )
            elif dtype in ["float64", "float32", "float16", "float128", "float256", "float"]:
                prop = FloatMetadataProperty(name=name, title=title, visible_for_annotators=self.visible_for_annotators)
            else:
                warnings.warn(f"Unhandled data type for column {col}: {dtype}")
                prop = None
            if prop is not None:
                properties.append(prop)
        return properties

    def _create_metadata_property_settings(
        self, dataset: Union[FeedbackDataset, RemoteFeedbackDataset], fields: List[str], overwrite: bool = False
    ) -> Union[FeedbackDataset, RemoteFeedbackDataset]:
        # Create a mock record with the specified fields
        records = [FeedbackRecord(fields={field: "mock the mocker" for field in fields})]
        # Extract text descriptives metrics from records
        extracted_metrics = self._extract_metrics_for_all_fields(records=records, fields=fields, overwrite=overwrite)
        # Cast integer and boolean columns to Python native types
        extracted_metrics = self._cast_to_python_types(extracted_metrics)
        # Clean column names
        extracted_metrics.columns = [self._clean_column_name(col) for col in extracted_metrics.columns]
        # Create metadata properties based on dataframe columns and data types
        metadata_properties = self._create_metadata_properties(extracted_metrics)
        # Add each metadata property iteratively to the dataset
        available_metadata_properties = [prop.name for prop in dataset.metadata_properties]
        for prop in metadata_properties:
            if prop.name in available_metadata_properties:
                # potentially overwrite visible_for_annotators settings
                metadata_property = dataset.metadata_property_by_name(prop.name)
                if overwrite and metadata_property.visible_for_annotators != prop.visible_for_annotators:
                    dataset.delete_metadata_property(prop.name)
                    dataset.add_metadata_property(prop)
            else:
                dataset.add_metadata_property(prop)
        return dataset

    def _add_text_descriptives_to_metadata(
        self, records: List[Union[FeedbackRecord, RemoteFeedbackRecord]], df: pd.DataFrame
    ) -> List[Union[FeedbackRecord, RemoteFeedbackRecord]]:
        """
        Add the text descriptives metrics extracted previously as metadata
        to a list of FeedbackDataset records.

        Args:
            records (List[Union[FeedbackRecord, RemoteFeedbackRecord]]): A list of FeedbackDataset or RemoteFeedbackDataset records.
            df (pd.DataFrame): The text descriptives dataframe.

        Returns:
            List[Union[FeedbackRecord, RemoteFeedbackRecord]]: A list of FeedbackDataset or RemoteFeedbackDataset records with extracted metrics added as metadata.
        """
        modified_records = []
        with Progress() as progress_bar:
            task = progress_bar.add_task(
                "Adding text descriptives to metadata...", total=len(records), visible=self.show_progress
            )
            for record, metrics in zip(records, df.to_dict("records")):
                filtered_metrics = {key: value for key, value in metrics.items() if not pd.isna(value)}
                filtered_metrics = {
                    key: int(value) if isinstance(value, float) and value.is_integer() else value
                    for key, value in filtered_metrics.items()
                }
                record.metadata.update(filtered_metrics)
                modified_records.append(record)
                progress_bar.update(task, advance=1)
        return modified_records

    def update_records(
        self,
        records: List[Union[FeedbackRecord, RemoteFeedbackRecord]],
        fields: Optional[List[str]] = None,
        overwrite: Optional[bool] = False,
    ) -> List[Union[FeedbackRecord, RemoteFeedbackRecord]]:
        """
        Extract text descriptives metrics from a list of FeedbackDataset or RemoteFeedbackDataset records,
        add them as metadata to the records and return the updated records.

        Args:
            records (List[Union[FeedbackRecord, RemoteFeedbackRecord]]): A list of FeedbackDataset or RemoteFeedbackDataset records.
            fields (List[str]): A list of fields to extract metrics for. If None, extract metrics for all fields.
            overwrite (Optional[bool]): Whether to overwrite existing metadata properties with the same name. Defaults to False.

        Returns:
            List[Union[FeedbackRecord, RemoteFeedbackRecord]]: A list of FeedbackDataset or RemoteFeedbackDataset records with text descriptives metrics added as metadata.

        Examples:
        >>> from argilla.client.feedback.integrations.textdescriptives import TextDescriptivesExtractor
        >>> records = [rg.FeedbackRecord(fields={"text": "This is a test."})]
        >>> tde = TextDescriptivesExtractor()
        >>> updated_records = tde.update_records(records)
        """
        available_fields = set()
        # ensure records are unique and don't update updated indices
        modified_records = []
        for rec in records:
            modified_records.append(rec)
            for key in rec.fields:
                available_fields.add(key)
        if fields is None:
            fields = list(available_fields)
        # Extract text descriptives metrics from records
        extracted_metrics = self._extract_metrics_for_all_fields(modified_records, overwrite=overwrite, fields=fields)
        # If the dataframe doesn't contain any columns, return the original records and log a warning
        if extracted_metrics.shape[1] == 0:
            warnings.warn(
                "No text descriptives metrics were extracted. This could be because the metrics contained NaNs."
            )
            return modified_records
        else:
            # Cast integer and boolean columns to Python native types
            extracted_metrics = self._cast_to_python_types(extracted_metrics)
            # Clean column names
            extracted_metrics.columns = [self._clean_column_name(col) for col in extracted_metrics.columns]
            # Add the metrics to the metadata of the records
            modified_records = self._add_text_descriptives_to_metadata(modified_records, extracted_metrics)
            return modified_records

    def update_dataset(
        self,
        dataset: Union[FeedbackDataset, RemoteFeedbackDataset],
        fields: Optional[List[str]] = None,
        update_records: Optional[bool] = True,
        overwrite: Optional[bool] = False,
    ) -> Union[FeedbackDataset, RemoteFeedbackDataset]:
        """
        Extract text descriptives metrics from records in a FeedbackDataset
        or RemoteFeedbackDataset, add them as metadata to the records and
        return the updated dataset.

        Args:
            dataset (Union[FeedbackDataset, RemoteFeedbackDataset]): A FeedbackDataset or RemoteFeedbackDataset.
            fields (Optional[List[str]]): A list of fields to extract metrics for. If None, extract metrics for all fields.
            update_records (Optional[bool]): Whether to update the records in the dataset too. Defaults to True.
            overwrite (Optional[bool]): Whether to overwrite existing metadata properties with the same name. Defaults to False.

        Returns:
            Union[FeedbackDataset, RemoteFeedbackDataset]: A FeedbackDataset or RemoteFeedbackDataset with text descriptives metrics added as metadata.

        Examples:
        >>> import argilla as rg
        >>> from argilla.client.feedback.integrations.textdescriptives import TextDescriptivesExtractor
        >>> dataset = rg.FeedbackDataset(...)
        >>> tde = TextDescriptivesExtractor()
        >>> updated_dataset = tde.update_dataset(dataset)
        """
        # If fields is None, use all fields
        if fields is None:
            fields = [field.name for field in dataset.fields]
        available_fields = [field.name for field in dataset.fields]
        if not all([field in available_fields for field in fields]):
            raise ValueError(f"Fields {fields} are not present in the dataset.")
        # Create metadata properties based on dataframe columns and data types
        dataset = self._create_metadata_property_settings(dataset=dataset, fields=fields, overwrite=overwrite)

        # Update the records in the dataset too
        if update_records:
            records = self.update_records(records=dataset.records, fields=fields, overwrite=overwrite)
            if isinstance(dataset, RemoteFeedbackDataset):
                dataset.update_records(records)
        return dataset



#  Copyright 2021-present, the Recognai S.L. team.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import warnings
from typing import List, Optional, Union

from argilla.client.feedback.dataset.local.dataset import FeedbackDataset
from argilla.client.feedback.dataset.remote.dataset import RemoteFeedbackDataset
from argilla.client.feedback.schemas.records import FeedbackRecord
from argilla.client.feedback.schemas.remote.records import RemoteFeedbackRecord
from argilla.client.feedback.schemas.vector_settings import VectorSettings
from argilla.utils.dependency import require_dependencies


class SentenceTransformersExtractor:
    """This class extracts a number of basic text descriptives from FeedbackDataset
    records using the TextDescriptives library and adds them as record metadata."""

    def __init__(
        self,
        model: Union["SentenceTransformer", str] = "TaylorAI/bge-micro-v2",
        show_progress: Optional[bool] = True,
        **kwargs,
    ):
        """
        Initialize the SentenceTransformersExtractor.

        Args:
            model (Optional[Union["SentenceTransformer", str]]): The SentenceTransformer model to use.
                Defaults to "TaylorAI/bge-micro-v2".
            show_progress (Optional[bool]): Whether to show the progress bar during encoding.
                Defaults to True.
            **kwargs: Additional keyword arguments to pass to the init of the SentenceTransformer model.

        Examples:
        >>> import argilla as rg
        >>> from argilla.client.feedback.integrations.textdescriptives import SentenceTransformersExtractor
        >>> ds = rg.FeedbackDataset(...)
        >>> tde = SentenceTransformersExtractor()
        >>> updated_ds = tde.update_dataset(ds)
        >>> updated_records = tde.update_records(ds.records)
        """
        require_dependencies("sentence_transformers")
        from sentence_transformers import SentenceTransformer

        self.show_progress = show_progress
        if isinstance(model, str):
            self.model = SentenceTransformer(model, **kwargs)
        else:
            if kwargs:
                warnings.warn(
                    "kwargs for initializing SentenceTransformer are ignored when passing a SentenceTransformer instance."
                )
            self.model = model
        self.embedding_dim = self.model.get_sentence_embedding_dimension()

    def _create_vector_settings(
        self,
        dataset: Union[FeedbackDataset, RemoteFeedbackDataset],
        fields: List[str],
        overwrite: bool = False,
    ) -> Union[FeedbackDataset, RemoteFeedbackDataset]:
        """
        Create or update vector settings for the given dataset.

        Args:
            dataset (Union[FeedbackDataset, RemoteFeedbackDataset]): The dataset to create or update vector settings for.
            fields (List[str]): The list of fields to create or update vector settings for.
            overwrite (bool, optional): Whether to overwrite existing vector settings if they have different dimensions. Defaults to False.

        Returns:
            Union[FeedbackDataset, RemoteFeedbackDataset]: The dataset with updated vector settings.
        """
        available_vector_settings = [setting.name for setting in dataset.vectors_settings]
        for field in fields:
            if field in available_vector_settings:
                setting = dataset.vector_settings_by_name(field)
                if overwrite:
                    if setting.dimensions != self.embedding_dim:
                        dataset.delete_vectors_settings(field)
                        dataset.add_vector_settings(
                            VectorSettings(
                                name=field,
                                dimensions=self.embedding_dim,
                            )
                        )
                else:
                    if setting.dimensions != self.embedding_dim:
                        raise ValueError(
                            f"Field {field} has a different embedding dimension ({dataset.vectors_settings[field].dimensions}) than the provided model ({self.embedding_dim})."
                        )
            else:
                dataset.add_vector_settings(
                    VectorSettings(
                        name=field,
                        dimensions=self.embedding_dim,
                    )
                )
        return dataset

    def _encode_single_field(
        self, records: List[Union[RemoteFeedbackRecord, FeedbackRecord]], field: str, overwrite: bool, **kwargs
    ) -> List[Union[FeedbackRecord, RemoteFeedbackRecord]]:
        """
        Encode a single field in the records with vectors.

        Args:
            records (List[Union[RemoteFeedbackRecord, FeedbackRecord]]): The list of records to encode.
            field (str): The field to encode.
            overwrite (bool): Whether to overwrite existing vectors for the field.
            **kwargs: Additional keyword arguments.

        Returns:
            List[Union[FeedbackRecord, RemoteFeedbackRecord]]: The updated list of records.
        """
        texts = []
        idxs = []
        # only
        for idx, record in enumerate(records):
            text = record.fields.get(field)
            if text:
                if field in record.vectors:
                    if overwrite:
                        texts.append(text)
                        idxs.append(idx)
                    else:
                        continue
                else:
                    texts.append(text)
                    idxs.append(idx)
        vectors = self.model.encode(texts, show_progress_bar=self.show_progress, **kwargs)
        for idx, vector in zip(idxs, vectors):
            records[idx].vectors[field] = vector.tolist()
        return records

    def update_records(
        self,
        records: List[Union[FeedbackRecord, RemoteFeedbackRecord]],
        fields: Optional[List[str]] = None,
        overwrite: Optional[bool] = False,
        **kwargs,
    ) -> List[Union[FeedbackRecord, RemoteFeedbackRecord]]:
        """
        Update the records by encoding fields with vectors.

        Args:
            records (List[Union[FeedbackRecord, RemoteFeedbackRecord]]): The records to update.
            fields (Optional[List[str]]): The fields to update. If None, use all available fields.
            overwrite (Optional[bool]): Whether to overwrite existing vectors. If None, use the value from
                the parent method. Defaults to None.
            **kwargs: Additional keyword arguments for encoding with sentence-transformers.

        Returns:
            List[Union[FeedbackRecord, RemoteFeedbackRecord]]: The updated records.

        Examples:
        >>> from argilla.client.feedback.integrations.sentencetransformers import SentenceTransformersExtractor
        >>> records = [rg.FeedbackRecord(fields={"text": "This is a test."})]
        >>> tde = SentenceTransformersExtractor()
        >>> updated_records = tde.update_records(records)
        """
        available_fields = set()
        # unwrap potential records from RemoteFeedbackDataset
        modified_records = []
        for rec in records:
            modified_records.append(rec)
            for key in rec.fields:
                available_fields.add(key)
        if fields is None:
            fields = list(available_fields)
        # encode fields
        for field in fields:
            modified_records = self._encode_single_field(modified_records, field, overwrite, **kwargs)
        return modified_records

    def update_dataset(
        self,
        dataset: Union[FeedbackDataset, RemoteFeedbackDataset],
        fields: Optional[List[str]] = None,
        update_records: Optional[bool] = True,
        overwrite: Optional[bool] = False,
        **kwargs,
    ) -> Union[FeedbackDataset, RemoteFeedbackDataset]:
        """
        Update the dataset by encoding fields with vectors.

        Args:
            dataset (Union[FeedbackDataset, RemoteFeedbackDataset]): The dataset to update.
            fields (Optional[List[str]]): The fields to update. If None, use all available fields.
            update_records (bool): Whether to update the records with vectors. Defaults to True.
            overwrite (bool): Whether to overwrite existing vectors. Defaults to False.
            **kwargs: Additional keyword arguments for encoding with sentence-transformers.

        Returns:
            Union[FeedbackDataset, RemoteFeedbackDataset]: The updated dataset.

        Examples:
        >>> import argilla as rg
        >>> from argilla.client.feedback.integrations.sentencetransformers import SentenceTransformersExtractor
        >>> dataset = rg.FeedbackDataset(...)
        >>> tde = SentenceTransformersExtractor()
        >>> updated_dataset = tde.update_dataset(dataset)
        """
        # If fields is None, use all fields
        if fields is None:
            fields = [field.name for field in dataset.fields]
        available_fields = [field.name for field in dataset.fields]
        if not all([field in available_fields for field in fields]):
            raise ValueError(f"Fields {fields} are not present in the dataset.")

        dataset = self._create_vector_settings(dataset=dataset, fields=fields, overwrite=overwrite)

        if update_records:
            records = self.update_records(records=dataset.records, fields=fields, overwrite=overwrite, **kwargs)
            if isinstance(dataset, RemoteFeedbackDataset):
                dataset.update_records(records)
        return dataset
